@@ -10,16 +10,16 @@ $builder->connect_signals();
 our %gui = (textbufferEditGenDescription         => $builder->get_object('textbufferEditGenDescription'),
             textbufferSnapshotDescription        => $builder->get_object('textbufferSnapshotDescription'),
             textbufferSnapshotDetailsDescription => $builder->get_object('textbufferSnapshotDetailsDescription'),
-            adjustmentEditDispVidMem             => $builder->get_object('adjustmentEditDispVidMem'),
-            adjustmentEditDispMonitor            => $builder->get_object('adjustmentEditDispMonitor'),
-            adjustmentEditDispQuality            => $builder->get_object('adjustmentEditDispQuality'),
-            adjustmentEditSysTimeOffset          => $builder->get_object('adjustmentEditSysTimeOffset'),
-            adjustmentEditSysMem                 => $builder->get_object('adjustmentEditSysMem'),
+            textbufferEditNetGeneric1            => $builder->get_object('textbufferEditNetGeneric1'),
+            textbufferEditNetGeneric2            => $builder->get_object('textbufferEditNetGeneric2'),
+            textbufferEditNetGeneric3            => $builder->get_object('textbufferEditNetGeneric3'),
+            textbufferEditNetGeneric4            => $builder->get_object('textbufferEditNetGeneric4'),
             adjustmentEditSysProcessor           => $builder->get_object('adjustmentEditSysProcessor'),
             adjustmentEditSysProcessorCap        => $builder->get_object('adjustmentEditSysProcessorCap'),
-            adjustmentEditSysLogoTime            => $builder->get_object('adjustmentEditSysLogoTime'),
+            adjustmentEditStorPortCount          => $builder->get_object('adjustmentEditStorPortCount'),
             appname                              => $builder->get_object('aboutdialog')->get_program_name(),
             appver                               => $builder->get_object('aboutdialog')->get_version(),
+            vboxEditIOPorts                      => $builder->get_object('vboxEditIOPorts'),
             websn                                => undef,
             pixbufstatuserror16                  => Gtk2::Gdk::Pixbuf->new_from_file("$Bin/share/remotebox/icons/status_error_16px.png"),
             pixbufotheros                        => Gtk2::Gdk::Pixbuf->new_from_file("$Bin/share/remotebox/icons/os/Other.png"));
@@ -53,100 +53,20 @@ our %signal = (fam          => $gui{comboboxNewOSFam}->signal_connect(changed =>
                netname2     => $gui{comboboxEditNetName2}->signal_connect(changed => \&net_name, $gui{checkbuttonEditNetEnable2}),
                netname3     => $gui{comboboxEditNetName3}->signal_connect(changed => \&net_name, $gui{checkbuttonEditNetEnable3}),
                netname4     => $gui{comboboxEditNetName4}->signal_connect(changed => \&net_name, $gui{checkbuttonEditNetEnable4}),
+               genericdrv1  => $gui{comboboxentryEditNetGenDriver1}->signal_connect(changed => \&net_generic_driver, $gui{checkbuttonEditNetEnable1}),
+               genericdrv2  => $gui{comboboxentryEditNetGenDriver2}->signal_connect(changed => \&net_generic_driver, $gui{checkbuttonEditNetEnable2}),
+               genericdrv3  => $gui{comboboxentryEditNetGenDriver3}->signal_connect(changed => \&net_generic_driver, $gui{checkbuttonEditNetEnable3}),
+               genericdrv4  => $gui{comboboxentryEditNetGenDriver4}->signal_connect(changed => \&net_generic_driver, $gui{checkbuttonEditNetEnable4}),
+               nameint1     => $gui{comboboxentryEditNetNameInt1}->signal_connect(changed => \&net_name_internal, $gui{checkbuttonEditNetEnable1}),
+               nameint2     => $gui{comboboxentryEditNetNameInt2}->signal_connect(changed => \&net_name_internal, $gui{checkbuttonEditNetEnable2}),
+               nameint3     => $gui{comboboxentryEditNetNameInt3}->signal_connect(changed => \&net_name_internal, $gui{checkbuttonEditNetEnable3}),
+               nameint4     => $gui{comboboxentryEditNetNameInt4}->signal_connect(changed => \&net_name_internal, $gui{checkbuttonEditNetEnable4}),
                stortype     => $gui{comboboxEditStorCtrType}->signal_connect(changed => \&storage_ctrtype));
 
-# Sensivity groups for enabling/disabling different parts of the gui
-our %sensgrp = (connect     => [$gui{menuitemNew},
-                                $gui{menuitemAdd},
-                                $gui{menuitemVMM},
-                                $gui{menuitemServerInfo},
-                                $gui{menuitemVBPrefs},
-                                $gui{toolbuttonNew},
-                                $gui{toolbuttonRefresh}],
-                running     => [$gui{menuitemAction},
-                                $gui{menuitemStop},
-                                $gui{menuitemPause},
-                                $gui{menuitemReset},
-                                $gui{menuitemKeyboard},
-                                $gui{menuitemDisplay},
-                                $gui{toolbuttonStop},
-                                $gui{toolbuttonReset},
-                                $gui{toolbuttonRemoteDisplay},
-                                $gui{buttonShowDetails},
-                                $gui{buttonRefreshSnapshots},
-                                $gui{buttonTakeSnapshot}],
-                saved       => [$gui{menuitemAction},
-                                $gui{menuitemStart},
-                                $gui{toolbuttonStart},
-                                $gui{buttonShowDetails},
-                                $gui{buttonRefreshSnapshots},
-                                $gui{buttonTakeSnapshot}],
-                poweroff    => [$gui{menuitemAction},
-                                $gui{menuitemStart},
-                                $gui{menuitemSettings},
-                                $gui{menuitemRemove},
-                                $gui{toolbuttonStart},
-                                $gui{toolbuttonSettings},
-                                $gui{buttonShowDetails},
-                                $gui{buttonRefreshSnapshots},
-                                $gui{buttonTakeSnapshot}],
-                paused      => [$gui{menuitemAction},
-                                $gui{menuitemResume},
-                                $gui{menuitemRemoteDisplay},
-                                $gui{toolbuttonRemoteDisplay},
-                                $gui{buttonShowDetails}],
-                unselected  => [$gui{menuitemAction},
-                                $gui{menuitemStart},
-                                $gui{menuitemStop},
-                                $gui{menuitemPause},
-                                $gui{menuitemResume},
-                                $gui{menuitemSettings},
-                                $gui{menuitemReset},
-                                $gui{menuitemRemove},
-                                $gui{menuitemKeyboard},
-                                $gui{menuitemDisplay},
-                                $gui{menuitemDVD},
-                                $gui{menuitemFloppy},
-                                $gui{toolbuttonStart},
-                                $gui{toolbuttonStop},
-                                $gui{toolbuttonSettings},
-                                $gui{toolbuttonRemoteDisplay},
-                                $gui{toolbuttonReset},
-                                $gui{buttonShowDetails},
-                                $gui{buttonRefreshSnapshots},
-                                $gui{buttonTakeSnapshot}],
-                stor_unsel  => [$gui{buttonEditStorAddAttach},
-                                $gui{buttonEditStorRemoveAttach},
-                                $gui{buttonEditStorRemoveCtr},
-                                $gui{comboboxEditStorCtrType},
-                                $gui{checkbuttonEditStorCache}],
-                ctr_sel_on  => [$gui{buttonEditStorAddAttach},
-                                $gui{buttonEditStorRemoveCtr},
-                                $gui{comboboxEditStorCtrType},
-                                $gui{checkbuttonEditStorCache}],
-                ctr_sel_off => [$gui{buttonEditStorAddAttach},
-                                $gui{buttonEditStorRemoveCtr},
-                                $gui{comboboxEditStorCtrType},
-                                $gui{checkbuttonEditStorCache}],
-                usb_sel     => [$gui{buttonEditUSBEdit},
-                                $gui{buttonEditUSBRemove},
-                                $gui{buttonEditUSBUp},
-                                $gui{buttonEditUSBDown}],
-                snap        => [$gui{buttonRestoreSnapshot},
-                                $gui{buttonDeleteSnapshot},
-                                $gui{buttonDetailsSnapshot}],
-                vbprefsnet  => [$gui{buttonVBPrefsDelNet},
-                                $gui{buttonVBPrefsEditNet}]);
-
 our (%hostspec, %chooser);
 
-#Glib->install_exception_handler (\&gui_ex_handler);
-
-# Simple GUI subs
-sub sensitive_on() { $_->set_sensitive(1) foreach (@_);}
-sub sensitive_off() { $_->set_sensitive(0) foreach (@_);}
-sub hide_on() { $_->hide foreach (@_); }
-sub hide_off() { $_->show foreach (@_); }
+# Work around a stupid bug in Glade 3.10.x that disables toolbutton menus
+$gui{toolbuttonStop}->set_menu($gui{menuStop});
 
 # Set the transient window's (ie parent window) sensitivity on
 sub transwin_sens_on() {
@@ -173,70 +93,120 @@ sub busy_window() {
     Gtk2->main_iteration() while Gtk2->events_pending();
 }
 
-sub handle_bioslogofilechooser() {
-    $gui{entryEditSysLogoPath}->set_text($_[0]);
-    &sys_bioslogopath();
+# Busy the pointer only
+sub busy_pointer() {
+    my ($window, $pointer) = @_;
+    if ($pointer) { eval { $window->window->set_cursor(Gtk2::Gdk::Cursor->new('watch')); }; }
+    else { eval{ $window->window->set_cursor(undef); }; }
+    Gtk2->main_iteration() while Gtk2->events_pending();
 }
 
-sub handle_machinefolderchooser() { $gui{entryVBPrefsGenDefMachineFolder}->set_text($_[0]); }
 
-sub handle_sharedfolderchooser() { $gui{entrySharedFolderPath}->set_text($_[0]); }
+sub handle_bioslogofilechooser() {
+    my ($location, $filearrayref) = @_;
+    my $file = ${$filearrayref}[0]->{FileName};
+    my $type = ${$filearrayref}[0]->{Type};
+
+    # We only care about the first file, there should only be one anyway
+    # double check there's a file, it's not .. and it's a bmp
+    if ($file and $file ne '..' and $type eq '.bmp') {
+        $gui{entryEditSysLogoPath}->set_text("$location/$file");
+        &sys_bioslogopath();
+    }
+}
+
+sub handle_machinefolderchooser() {
+    my ($location, $filearrayref) = @_;
+    my $file = ${$filearrayref}[0]->{FileName};
+    my $type = ${$filearrayref}[0]->{Type};
+
+    # Depending on what the user does, there will either be an additional directory
+    # to append or not. Type (Dir) already excludes (Parent)
+    $location .= "/$file" if ($file and $type eq '(Dir)');
+    $gui{entryVBPrefsGenDefMachineFolder}->set_text($location) if ($location);
+}
+
+sub handle_sharedfolderchooser() {
+    my ($location, $filearrayref) = @_;
+    my $file = ${$filearrayref}[0]->{FileName};
+    my $type = ${$filearrayref}[0]->{Type};
+
+    # Depending on what the user does, there will either be an additional directory
+    # to append or not. Type (Dir) already excludes (Parent)
+    $location .= "/$file" if ($file and $type eq '(Dir)');
+    $gui{entrySharedFolderPath}->set_text($location) if ($location);
+}
 
 sub handle_vboxfilechooser() {
-    my ($vbox) = @_;
-    my $IMachine = IVirtualBox_openMachine($gui{websn}, $vbox);
-    if ($IMachine) {
-        IVirtualBox_registerMachine($gui{websn}, $IMachine);
-        &addrow_log("Added guest from $vbox.");
-        &fill_list_guest();
+    my ($location, $filearrayref) = @_;
+    my $file = ${$filearrayref}[0]->{FileName};
+    my $type = ${$filearrayref}[0]->{Type};
+
+    # We only care about the first file, there should only be one anyway
+    # double check there's a file, it's not .. and it's a vbox
+    if ($file and $file ne '..' and $type eq '.vbox') {
+        my $IMachine = IVirtualBox_openMachine($gui{websn}, "$location/$file");
+        if ($IMachine) {
+            IVirtualBox_registerMachine($gui{websn}, $IMachine);
+            &addrow_log("Imported guest from $location/$file");
+            &fill_list_guest();
+        }
+        else { &addrow_log("Failed to import guest from $location/$file"); }
     }
-    else { &addrow_log("Failed to add guest from $vbox."); }
 }
 
+# Sets up the file chooser for selecting a BMP files
 sub show_bioslogofilechooser {
     my ($widget) = @_;
     $gui{dialogRemoteFileChooser}->set_transient_for($gui{dialogEdit});
     $gui{dialogRemoteFileChooser}->set_title("Choose BIOS logo on $endpoint");
     my $basedir = $gui{entryEditSysLogoPath}->get_text();
     if ($basedir) { (undef, $basedir, undef) = fileparse($basedir); }
-    $chooser{ext} = 'bmp';
+    $chooser{filter} = '^.*\.bmp$'; # Only select .bmp files
     $chooser{handler} = \&handle_bioslogofilechooser;
-    $chooser{populator} = \&fill_list_remotefiles_ext;
+    $chooser{selector} = 'GTK_SELECTION_SINGLE';
     &show_remotefilechooser_window($basedir);
 }
 
+# Sets up the file chooser for selecting a VBOX file
 sub show_vboxfilechooser {
     my ($widget) = @_;
     $gui{dialogRemoteFileChooser}->set_transient_for($gui{windowMain});
     $gui{dialogRemoteFileChooser}->set_title("Choose guest to add on $endpoint");
-    $chooser{ext} = 'vbox';
+    $chooser{filter} = '^.*\.vbox$'; # Only select .vbox files
     $chooser{handler} = \&handle_vboxfilechooser;
-    $chooser{populator} = \&fill_list_remotefiles_ext;
+    $chooser{selector} = 'GTK_SELECTION_SINGLE';
     &show_remotefilechooser_window();
 }
 
+# Sets up the file chooser for selecting a medium (or any file actually)
 sub show_vmmfilechooser() {
     my ($widget) = @_;
     $gui{dialogRemoteFileChooser}->set_transient_for($gui{dialogVMM});
     $gui{dialogRemoteFileChooser}->set_title("Choose Media Image on $endpoint");
+    $chooser{filter} = '.*'; # Allow any file in the list
     $chooser{handler} = \&vmm_add;
-    $chooser{populator} = \&fill_list_remotefiles;
+    $chooser{selector} = 'GTK_SELECTION_MULTIPLE';
     &show_remotefilechooser_window();
 }
 
+# Sets up the file chooser for selecting the default machine directory
 sub show_machinefolderchooser() {
     $gui{dialogRemoteFileChooser}->set_transient_for($gui{dialogVBPrefs});
     $gui{dialogRemoteFileChooser}->set_title("Choose Machine Folder on $endpoint");
+    $chooser{filter} = ' ^'; # Effectively won't match any files, listing directories only
     $chooser{handler} = \&handle_machinefolderchooser;
-    $chooser{populator} = \&fill_list_remotefolders;
+    $chooser{selector} = 'GTK_SELECTION_SINGLE';
     &show_remotefilechooser_window($gui{entryVBPrefsGenDefMachineFolder}->get_text());
 }
 
+# Sets up the file chooser for selecting a shared folder
 sub show_sharedfolderchooser() {
     $gui{dialogRemoteFileChooser}->set_transient_for($gui{dialogShared});
     $gui{dialogRemoteFileChooser}->set_title("Choose Shared Folder on $endpoint");
+    $chooser{filter} = ' ^'; # Effectively won't match any files, listing directories only
     $chooser{handler} = \&handle_sharedfolderchooser;
-    $chooser{populator} = \&fill_list_remotefolders;
+    $chooser{selector} = 'GTK_SELECTION_SINGLE';
     &show_remotefilechooser_window($gui{entrySharedFolderPath}->get_text());
 }
 
@@ -246,29 +216,30 @@ sub show_remotefilechooser_window() {
     &transwin_sens_off($gui{dialogRemoteFileChooser});
     my $IAppliance = IVirtualBox_createAppliance($gui{websn});
     $gui{IVFSExplorer} = IAppliance_createVFSExplorer($IAppliance, "file://$basedir");
-    &{$chooser{populator}}($basedir);
+    &fill_list_remotefiles($basedir);
+    $gui{treeviewRemoteFileChooser}->get_selection->set_mode($chooser{selector}) if ($chooser{selector});
     my $response = $gui{dialogRemoteFileChooser}->run();
     $gui{dialogRemoteFileChooser}->hide();
     IManagedObjectRef_release($gui{IVFSExplorer});
     IManagedObjectRef_release($IAppliance);
 
     if ($response eq 'ok') {
-        my $fileref = &getsel_list_remotefiles();
+        my $filearrayref = &getsel_list_remotefiles();
         my $location = $gui{entryRemoteFileChooserLocation}->get_text();
-        $location = "$location/$$fileref{FileName}" if ($$fileref{FileName});
-        &{$chooser{handler}}($location);
+        &{$chooser{handler}}($location, $filearrayref, $chooser{userdata});
     }
 
     &transwin_sens_on($gui{dialogRemoteFileChooser});
 }
 
-sub refresh_remotefilechooser() { &{$chooser{populator}}($gui{entryRemoteFileChooserLocation}->get_text()); }
+sub refresh_remotefilechooser() { &fill_list_remotefiles($gui{entryRemoteFileChooserLocation}->get_text()); }
 
 sub cdup_remotefilechooser() {
     IVFSExplorer_cdUp($gui{IVFSExplorer});
-    &{$chooser{populator}}(IVFSExplorer_getPath($gui{IVFSExplorer}));
+    &fill_list_remotefiles(IVFSExplorer_getPath($gui{IVFSExplorer}));
 }
 
+# Display a progress window for tasks which can take a long time
 sub show_progress_window() {
     my ($IProgress, $msg, $cancel, $transwin) = @_;
     my $resultcode = 0;
@@ -278,8 +249,8 @@ sub show_progress_window() {
     $gui{dialogProgress}->set_transient_for($transwin) if ($transwin);
     $gui{progressbar}->set_fraction(0);
 
-    if ($cancel) { &hide_off($gui{buttonProgressCancel}); }
-    else { &hide_on($gui{buttonProgressCancel}); }
+    if ($cancel) { $gui{buttonProgressCancel}->show(); }
+    else { $gui{buttonProgressCancel}->hide(); }
 
     $timer = Glib::Timeout->add(1000,
         sub {
@@ -306,6 +277,7 @@ sub show_progress_window() {
     return $resultcode;
 }
 
+# Save RemoteBox's window position for restoration later
 sub save_window_pos() {
     my ($winname) = @_;
     my $alloc = $gui{$winname}->allocation;
@@ -315,6 +287,7 @@ sub save_window_pos() {
     &rbprefs_save();
 }
 
+# Restore RemoteBox's window position to the last save position
 sub restore_window_pos() {
     my ($winname) = @_;
 
@@ -354,48 +327,130 @@ sub show_progress_window2() {
     Glib::Source->remove($timer) if ($timer);
 }
 
+# Permit certain chars only for a guest name
 sub validate_name() {
     my ($entry, $char, $len, $pos) = @_;
     $char =~ s/[\?\/\;\*\\\<\>\|\.]//; # Strip these chars
     return $char, $pos;
 }
 
+# Permit only hexadecimal in an entry
 sub validate_hex() {
     my ($entry, $char, $len, $pos) = @_;
     $char =~ s/[^A-F0-9a-f]//; # Strip everything but these chars
     return $char, $pos;
 }
 
+# Permit only valid numbers for a port entry
 sub validate_port() {
     my ($entry, $char, $len, $pos) = @_;
     $char =~ s/[^0-9,-]//; # Strip everything but these chars
     return $char, $pos;
 }
 
+# Permit only numbers in an entry
 sub validate_number() {
     my ($entry, $char, $len, $pos) = @_;
     $char =~ s/[^0-9]//; # Strip everything but these chars
     return $char, $pos;
 }
 
+# Permit only valid chars in an IPV4 entry
 sub validate_ip() {
     my ($entry, $char, $len, $pos) = @_;
     $char =~ s/[^0-9,.]//; # Strip everything but these chars
     return $char, $pos;
 }
 
+# Permit only valid chars in an IPV6 entry
 sub validate_ipv6() {
     my ($entry, $char, $len, $pos) = @_;
     $char =~ s/[^A-F0-9a-f,:]//; # Strip everything but these chars
     return $char, $pos;
 }
 
-sub gui_ex_handler {
-    &show_err_msg('guiexception', $gui{messagedialogError}, "This appears to be a bug in RemoteBox. Unfortunately RemoteBox will be terminated when the dialog closes:\n\n @_");
-    &quit_remotebox();
-    0; # Remove after running once, we've probably quit before here anyway
+sub combobox_set_active_text {
+    my ($combobox, $text) = @_;
+    my $i = -1;
+    $combobox->get_model->foreach (
+                            sub {
+                                my ($model, $path, $iter) = @_;
+                                if ($text eq $model->get_value ($iter, 0)) {
+                                    ($i) = $path->get_indices;
+                                    return 1; # stop
+                                }
+                                return 0; # continue
+                            }
+                          );
+    $combobox->set_active($i);
 }
 
+# Adds appropriate units to a spinbox when specifying memory or disk
+sub spinbox_units() {
+    my ($widget) = @_;
+    my $adjustment = $widget->get_adjustment();
+    my $adjval = int($adjustment->get_value());
+    my $txt;
 
+    if ($adjval < 1024) {
+        $adjustment->step_increment(1.00);
+        $txt = "$adjval MB";
+    }
+    elsif ($adjval < 1048576) {
+        $adjustment->step_increment(10.25);
+        $adjval /= 1024;
+        $txt = sprintf('%0.2f GB', $adjval);
+    }
+    else {
+        $adjustment->step_increment(10486);
+        $adjval /= 1048576;
+        $txt = sprintf('%0.2f TB', $adjval);
+    }
+
+    $widget->set_text($txt);
+    return 1;
+}
+
+# Adds appropriate units to a spinbox when specifying time
+sub spinbox_units_time() {
+    my ($widget) = @_;
+    my $adjustment = $widget->get_adjustment();
+    my $adjval = int($adjustment->get_value());
+    my $txt;
+
+    if ($adjval < 1000) {
+        $adjustment->step_increment(1.00);
+        $txt = "$adjval ms";
+    }
+    elsif ($adjval < 60000) {
+        $adjustment->step_increment(10.00);
+        $adjval /= 1000;
+        $txt = sprintf('%0.2f secs', $adjval);
+    }
+    else {
+        $adjustment->step_increment(600.00);
+        $adjval /= 60000;
+        $txt = sprintf('%0.2f mins', $adjval);
+    }
+
+    $widget->set_text($txt);
+    return 1;
+}
+
+# Adds appropriate units to a spinbox when specifying percent
+sub spinbox_units_pc() {
+    my ($widget) = @_;
+    my $adjustment = $widget->get_adjustment();
+    my $adjval = int($adjustment->get_value());
+    $widget->set_text($adjval . '%');
+    return 1;
+}
+
+# Callback set on a timer to attempt to keep the connection alive, even in the
+# case where a user has set a timeout. This callback needs to be cheap
+sub heartbeat() {
+    IVirtualBox_getVersion($gui{websn}) if ($gui{websn});
+    return 1; # Return 1 to stop the timer from being removed
+}
 
 1;
